// ----------------------------------------------------------------------------
// VFIS 
// ----------------------------------------------------------------------------
// Verified Fredholm Integral Equation Solver
// for linear Fredholm Integral Equations of the second kind
// (Parallel System Version)
// ----------------------------------------------------------------------------
// Verifizierter Integralgleichungsloeser 
// fuer lineare Fredholmsche Integralgleichungen zweiter Art
// (Parallele Systemversion)
// ----------------------------------------------------------------------------
// Copyright (c) 2004-2007 Markus Grimmer
//
// markus.grimmer@math.uni-wuppertal.de
// ----------------------------------------------------------------------------

// This file is part of the Verified Fredholm Integral Equation Solver VFIS.
// 
// VFIS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// VFIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with VFIS; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

/**
 * @file System-FIGL-Loeser-par-Bsp.cpp
 * Example Application file for VFIS Integral Equation Solver.
 *
 * The application is designed to run on the parallel computer
 * ALiCEnext; the handling of input parameters and output files
 * works an ALiCEnext, but you might have to make a few changes
 * on other parallel machines.
 *
 * This application takes the following command line parameters:
 *
 * <ul>
 * <li> 1 Taylor order </li>
 * <li> 2 String representing the Taylor order (for file name creation)</li>
 * <li> 3 System order</li>
 * <li> 4 String representing the system order (for file name creation)</li>
 * <li> 5 Directory name for local file creation on ALiCEnext nodes</li>
 * <li> 6 Directory name as final output file destination</li>
 * </ul>
 */

#include <mpi.h>
#include <string>
#include <iostream>    
#include <fstream>    
#include <sstream>
#include <cstdlib>

#include "System-FIGL-Loeser-par.hpp"

//
// This is the file you should put the integral equation's 
// kernel function and right hand side into:
//
#include "functions.in" 

/**
 * Typedef for kernel function pointer.
 */
typedef dim2taylor (*KFkt)(dim2taylor_vector&);
/**
 * Typedef for right hand side function pointer.
 */
typedef itaylor (*ReFkt)(itaylor&);

// ----------------------------------------------------------------------------

/**
 * Executes a test run. For multiple runs, put a series of calls in your application.
 *
 * @param[in] ofbasename Prefix of output file name
 * @param[in] ord        Taylor order
 * @param[in] ordstr     String representing the Taylor order (for file name creation)
 * @param[in] sysord     System order
 * @param[in] sysordstr  String representing the system order (for file name creation)
 * @param[in] K          Kernel function
 * @param[in] KS         Domain of first variable of kernel function
 * @param[in] KT         Domain of second variable of kernel function
 * @param[in] ReS        Right hand side function
 * @param[in] GS         Domain of riht hand side 
 * @param[in] procs      Number of parallel processes
 * @param[in] mypid      Process ID
 * @param[out] commerrc  Error code
 * @param[in] datadir    Directory name for local file creation on ALiCEnext nodes
 * @param[in] targetdir  Directory name as final output file destination
 */
int executetest(string ofbasename, int ord, string ordstr, 
                int sysord, string sysordstr, 
                KFkt K, interval KS, interval KT,
                ReFkt ReS, interval GS, int procs, int mypid,
                int& commerrc, string datadir, string targetdir)
{
  int width=23, digits=18;

  // ------------------------------------------------------------------------
  // Determine and open Ouput file 
  // ------------------------------------------------------------------------

      ofstream ausg;
      ostringstream ausgdats, plotdats;
      string ausgabe, ausgabedatei, plotdatei;

      ausgabe=ofbasename+"-t"+ordstr+"-s"+sysordstr;

      // *************************************************************************

      //To make a new directory insert the following:
      
        string syscmd1= "if test -d "
                        +datadir+"; then echo \"Verzeichnis vorhanden\"; else mkdir "
                        +datadir+"; fi;";
        system(syscmd1.c_str());
      
      //(end insert)
 
      ausgdats << datadir << "/" << ausgabe << "." << procs << "p." 
               << mypid << ".out";
      plotdats << datadir << "/" << ausgabe << "." << procs << "p." 
               << mypid << ".mpl";
      ausgabedatei=ausgdats.str();
      plotdatei=plotdats.str();

  // ---------------------------------------------------------------------------
  // Define Integral Equation
  // ---------------------------------------------------------------------------

  IntegralGleichungsSystem IGLSys;

  int root=0;
  if (mypid==root)
  {
    IGLSys=IntegralGleichungsSystem(K,KS,KT,ReS,GS,sysord,ord,ausg,ausgabedatei);  
    
    if (commerrc=IGLSys.DistributeValues(root)!=MPI_SUCCESS)
       MPI_Fehler(commerrc,"MPI_Bcast", 
                  "IntegralGleichungsSystem::DistributeValues",mypid);
  }
  else
  {
    IGLSys.PartInitIntegralGleichungsSystem(K,KS,KT,ReS,GS,
                                            sysord,ord,ausg,ausgabedatei);  
    if (commerrc=IGLSys.DistributeValues(root)!=MPI_SUCCESS)
       MPI_Fehler(commerrc,"MPI_Bcast",
                  "IntegralGleichungsSystem::DistributeValues",mypid);
  }
  
  ausg.open(ausgabedatei.c_str(),ios_base::app);

  clock_t t;      
  start_clock(t); 
  
  cout << "Starte Berechnung der Einschliessung..." << endl;
  ausg << "Starte Berechnung der Einschliessung..." << endl;

  ausg.close();

  // ------------------------------------------------------------------------
  // Compute enclosure of solution
  // ------------------------------------------------------------------------

  IGLSys.Einschliessung_berechnen(ausg,ausgabedatei,procs,mypid,commerrc);

  ausg.open(ausgabedatei.c_str(),ios_base::app);
  
  print_time_used(t,ausg);

  
  // ------------------------------------------------------------------------

  if (commerrc!=0)
    MPI_Fehler(commerrc,"(MATINV Communication)", "MatInv",-1);

  // ------------------------------------------------------------------------
  // Export result to Maple code and move file to target directory
  // ------------------------------------------------------------------------

  if (mypid==root)
  {
    IGLSys.Export_Plot("Maple",plotdatei,8);
    string syscmd_mvplotf= "mv "+ plotdatei + " " + targetdir;
    system(syscmd_mvplotf.c_str());
  }

  cout << flush << "Ende." << endl;
  ausg << flush << "Ende." << endl;
 
  // ------------------------------------------------------------------------
  // Close output file and move file to target directory
  // ------------------------------------------------------------------------

  ausg.close();
      
  string writehostcmd="hostname >> "+ausgabedatei;
  system(writehostcmd.c_str());
  
  string syscmd_mvres= "mv "+ ausgabedatei + " " + targetdir;
  system(syscmd_mvres.c_str());

  return 0;
}

// ----------------------------------------------------------------------------

/**
 * Main function processing the command line parameters, initializing the
 * parallel environment and calling the test execution function.
 * 
 * @param[in] argc Argument count
 * @param[in] argv Argument vector
 */
int main(int argc, char *argv[]) 
{

  // ---------------------------------------------------------------------------
  // MPI Communication Init
  // ---------------------------------------------------------------------------

  int procs;
  int mypid;
  int commerrc;
  MPI_Status status;

  if ((commerrc=MPI_Init(&argc, &argv)) != MPI_SUCCESS)
    MPI_Fehler(commerrc,"MPI_Init", "main",-1);
    
  if ((commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid)) != MPI_SUCCESS)
    MPI_Fehler(commerrc,"MPI_Comm_rank", "main",mypid);
    
  if ((commerrc=MPI_Comm_size(MPI_COMM_WORLD, &procs)) != MPI_SUCCESS)
    MPI_Fehler(commerrc,"MPI_Comm_size", "main",procs);

  // Define MPI-CXSC-Data types:
  MPI_Define_CXSC_Types();

  if (argc<7)
  {
    cerr << "Nicht genuegend Argumente. Abbruch." << endl; 
  }
  else
  {
    int arg_ord=atoi(argv[1]);
    string arg_ordstr(argv[2]);
    int arg_sysord=atoi(argv[3]);
    string arg_sysordstr(argv[4]);

    // -------------------------------------------------------------------------
    // Output file directories
    // -------------------------------------------------------------------------

    ostringstream datadirs, targetdirs;
    string datadir, targetdir;

    datadirs << argv[5];
    targetdirs << argv[6];
    datadir = datadirs.str();
    targetdir = targetdirs.str();

    // -------------------------------------------------------------------------

    executetest("Bsp",arg_ord,arg_ordstr,arg_sysord,arg_sysordstr,
                K,KDefS,KDefT,ReSeite,ReSeiteDef,procs,mypid,
                commerrc,datadir,targetdir);
  }

  // ------------------------------------------------------------------------
   
  if ((commerrc=MPI_Finalize()))
     MPI_Fehler(commerrc,"MPI_Finalize","main",mypid);

  return 0;
  
}
